# %% 
# 0. The basics
import numpy as np
import matplotlib
from matplotlib import pyplot as plt

# %% 
# ===========================================================================
# 1. Generate the truth (without transient)
from misc_funs import natrun
Nx = 20
tmax = 4; dt = 0.025
x,t,ut,ug0 = natrun(Nx,tmax)
Nsteps = np.size(t); print 'Nsteps',Nsteps

# Plot the trajectory
nrow=4; ncol=5

plt.figure()
for jpl in range(Nx):
 plt.subplot(nrow,ncol,jpl+1)
 plt.plot(t,ut[:,jpl],'k')
 plt.xlabel('time')
 plt.grid(True)
del jpl


#%%
# ========================================================================
# 2. Observations
from miscobs import getHR, genobs
# Select the period of observations (in model steps) and the grid
# Gridobs: 1: all, 2: on/off, 3: half, 4: every 4, 5: single obs
period_obs = 2
gridobs = 2 
stdobs = 1

Nx_obs, loc_obs, H, R, Rsq, invR = getHR(gridobs,Nx,stdobs)

loc_nobs = np.setdiff1d(x,loc_obs)
locs = [loc_obs,loc_nobs]
# y
tobs,yobs = genobs(dt,ut,Nsteps,Nx_obs,H,period_obs,Rsq)

# plot
plt.figure()
for jpl in range(Nx):
 plt.subplot(nrow,ncol,jpl+1)
 plt.plot(t,ut[:,jpl],'k')
 plt.xlabel('time')
 plt.grid(True)
del jpl
for jpl in range(Nx_obs):
 plt.subplot(nrow,ncol,loc_obs[jpl]+1)
 plt.autoscale(False) 
 plt.scatter(tobs,yobs[:,jpl],20,'r')
del jpl


# %%
#===========================================================================
# 3. Variational data assimilation
from rmse_spread import rmse_spread 
from misc_funs import getBc
Bc,Bc_sq = getBc(Nx)
# plot
plt.figure()
lim = 2
my_cmap = matplotlib.cm.get_cmap('BrBG_r')
plt.imshow(np.array(Bc),interpolation="nearest",cmap=my_cmap,vmin=-lim,vmax=lim)
plt.colorbar()


# %%
# --------------------------------------------------------------
# 3.1. As a base for coparison, do a simple 3DVar and 4DVar
from var3dfile import var3d
from var4dfile import var4d

ub3,ua3 = var3d(ug0,t,x,H,yobs,period_obs,gridobs,Bc_sq,invR)        
obsperwin = 2
ub4,ua4 = var4d(ug0,t,x,H,yobs,period_obs,obsperwin,gridobs,Bc_sq,invR)

rmseb3 = np.empty((Nsteps,2));     rmsea3 = np.empty((Nsteps,2));     
rmseb4 = np.empty((Nsteps,2));     rmsea4 = np.empty((Nsteps,2));

for job in range(2):
 rmseb3[:,job] = rmse_spread(ut[:,locs[job].astype(int)], \
                ub3[:,locs[job].astype(int)],None,1)
 rmsea3[:,job] = rmse_spread(ut[:,locs[job].astype(int)], \
                ua3[:,locs[job].astype(int)],None,1)
 rmseb4[:,job] = rmse_spread(ut[:,locs[job].astype(int)], \
                ub4[:,locs[job].astype(int)],None,1)
 rmsea4[:,job] = rmse_spread(ut[:,locs[job].astype(int)], \
                ua4[:,locs[job].astype(int)],None,1)
del job

plt.figure()
for jpl in range(Nx):
 plt.subplot(nrow,ncol,jpl+1)
 plt.plot(t,ut[:,jpl],'-k')
 plt.plot(t,ub3[:,jpl],'-c')
 plt.plot(t,ua3[:,jpl],'-m')
 plt.plot(t,ub4[:,jpl],'-b')
 plt.plot(t,ua4[:,jpl],'-r')
 plt.xlabel('time')
 plt.title('x['+str(jpl+1)+']')
 plt.grid(True)
del jpl
for jpl in range(Nx_obs):
 plt.subplot(nrow,ncol,loc_obs[jpl]+1)
 plt.autoscale(False) 
 plt.scatter(tobs,yobs[:,jpl],20,'k')
del jpl


#%%
# 3.2. Compare 3DVar and SC-4DVar
plt.figure()
for job in range(2):
 plt.subplot(1,2,job+1)
 plt.plot(t,rmseb3[:,job],'-c.',label='bgd-3DVar')
 plt.plot(t,rmsea3[:,job],'-m.',label='ana-3DVar')
 plt.plot(t,rmseb4[:,job],'-b.',label='bgd-4DVar')
 plt.plot(t,rmsea4[:,job],'-r.',label='ana-4DVar')
 plt.legend()
del job


# %%
#=============================================================================
# 4. Ensemble data assimilation
from etkf16 import getlocmat
lam = 2;  loctype = 1 #(Gaspari-Cohn)
Lxx = getlocmat(Nx,Nx,np.eye(Nx),lam,loctype) # get the localisation matrix
Lxy = getlocmat(Nx,Nx_obs,H,lam,loctype) # get the localisation matrix
loc_cmap = matplotlib.cm.get_cmap('gray_r')

plt.figure()
plt.subplot(1,2,1)
plt.pcolor(np.flipud(Lxx),cmap=loc_cmap)
plt.subplot(1,2,2)
plt.pcolor(np.flipud(Lxy),cmap=loc_cmap)
plt.colorbar()


# %%
# ----------------------------------------------------------------------------
# 4.1 Do a LETKF and analyse the ensemble covariances
from etkf16 import etkf_l96
M = 8
Ubkf,ubkf,Uakf,uakf = etkf_l96(ug0,t,x,M,Nx_obs,H,R,yobs,period_obs,lam,Lxy)

plt.figure()
for jpl in range(Nx):
 plt.subplot(nrow,ncol,jpl+1)
 plt.plot(t,ut[:,jpl],'-k')
 plt.plot(t,Ubkf[:,jpl,:],'-c')
 plt.plot(t,Uakf[:,jpl,:],'-m')
 plt.plot(t,ubkf[:,jpl],'-b')
 plt.plot(t,uakf[:,jpl],'-r')
 plt.grid(True)
del jpl
for jpl in range(Nx_obs):
 plt.subplot(nrow,ncol,loc_obs[jpl]+1)
 plt.autoscale(False) 
 plt.scatter(tobs,yobs[:,jpl],20,'k')
del jpl

rmsebkf = np.empty((Nsteps,2));     rmseakf = np.empty((Nsteps,2));
for job in range(2):
 rmsebkf[:,job] = rmse_spread(ut[:,locs[job].astype(int)], \
                ubkf[:,locs[job].astype(int)],None,1)
 rmseakf[:,job] = rmse_spread(ut[:,locs[job].astype(int)], \
                uakf[:,locs[job].astype(int)],None,1)
del job

plt.figure()
for job in range(2):
 plt.subplot(1,2,job+1)
 plt.plot(t,rmsebkf[:,job],'-c.')
 plt.plot(t,rmseakf[:,job],'-m.')
del job


#%%
# 4.2. Do some comparisons in terms of RMSE
plt.figure()
for job in range(2):
 plt.subplot(1,2,job+1)
 plt.plot(t,rmseb3[:,job],'-c.',label='bgd-3DVar')
 plt.plot(t,rmsea3[:,job],'-m.',label='ana-3DVar')
 plt.plot(t,rmseb4[:,job],'-b.',label='bgd-4DVar')
 plt.plot(t,rmsea4[:,job],'-r.',label='ana-4DVar')
 plt.plot(t,rmsebkf[:,job],'-y.',label='bgd-LETKF')
 plt.plot(t,rmseakf[:,job],'-g.',label='ana-LETKF')
 plt.legend()
del job


#%%
# ----------------------------------------------------------------------
# 5. Now let's compare the climatological Bc with some sample Bc's
nsample = 3
ind = np.arange(period_obs,(nsample+1)*period_obs,period_obs)
Pbs_kf = np.empty((Nx,Nx,nsample))
LPbs_kf = np.empty((Nx,Nx,nsample))
for j in range(nsample):
 aux = np.squeeze(Ubkf[ind[j],:,:])
 aux = np.cov(aux,ddof=1)
 Pbs_kf[:,:,j] = aux
 LPbs_kf[:,:,j] = Lxx*aux
del j

plt.figure()
lim = 1
my_cmap = matplotlib.cm.get_cmap('BrBG_r')
plt.subplot(2,nsample+1,1)
plt.imshow(np.array(Bc),interpolation="nearest",cmap=my_cmap,vmin=-lim,vmax=lim)
plt.colorbar()   

plt.subplot(2,nsample+1,nsample+2)
plt.imshow(np.array(Bc),interpolation="nearest",cmap=my_cmap,vmin=-lim,vmax=lim)
plt.colorbar()   

for j in range(nsample):
 plt.subplot(2,nsample+1,2+j)   
 plt.imshow(np.array(Pbs_kf[:,:,j]),interpolation="nearest",cmap=my_cmap,vmin=-lim,vmax=lim)
 plt.colorbar()
 plt.subplot(2,nsample+1,nsample+3+j)   
 plt.imshow(np.array(LPbs_kf[:,:,j]),interpolation="nearest",cmap=my_cmap,vmin=-lim,vmax=lim)
 plt.colorbar()   
del j


#%%
###########################################################################
# 6 Hybrid DA 
lam = 2; loctype = 1
Lxx = getlocmat(Nx,Nx,np.eye(Nx),lam,loctype) 
Lxy = getlocmat(Nx,Nx_obs,H,lam,loctype) 

import h4Dkf;   reload(h4Dkf)
from h4Dkf import etkf4DVar
loch = 1


#%% 
# 6.1 Hybrid DA part 1
beta = [0.6,0.4]
obsperwin = 2

ubh,uah,Uaenh,uaenh = etkf4DVar(ug0,t,x,R,invR,H,yobs,period_obs,obsperwin,\
                           gridobs,Nx_obs,Bc,Bc_sq,lam,Lxx,Lxy,loch,M,beta)

plt.figure()
for jpl in range(Nx):
 plt.subplot(nrow,ncol,jpl+1)
 plt.plot(t,ut[:,jpl],'-k')
 plt.plot(t,ubh[:,jpl],'-b')
 plt.plot(t,uah[:,jpl],'-r')
 plt.plot(t,Uaenh[:,jpl,:],'-y')
 plt.plot(t,uaenh[:,jpl],'-g')
 plt.grid(True)
del jpl
for jpl in range(Nx_obs):
 plt.subplot(nrow,ncol,loc_obs[jpl]+1)
 plt.autoscale(False) 
 plt.scatter(tobs,yobs[:,jpl],20,'r')
del jpl

rmsebh = np.empty((Nsteps,2));     rmseah = np.empty((Nsteps,2));

for job in range(2):
 rmsebh[:,job] = rmse_spread(ut[:,locs[job].astype(int)], \
                np.squeeze(ubh[:,locs[job].astype(int)]),None,1)
 rmseah[:,job] = rmse_spread(ut[:,locs[job].astype(int)], \
                np.squeeze(uah[:,locs[job].astype(int)]),None,1)
del job

plt.figure()
for job in range(2):
 plt.subplot(1,2,job+1)
 plt.plot(t,rmsebh[:,job],'-c.')
 plt.plot(t,rmseah[:,job],'-m.')
del job


#%%
##############################################################################
# 4. Hybrid DA part 2: Avoiding the TLM and adjoint
# 4.1, Evolution of covariances by two ways
from transmat import transmat_l96
from misc_funs import evolcov, covfamrun

uref, tmat, seed = transmat_l96(ug0,t,x)
lags = 30
Bt, B0t = evolcov(Bc,tmat,Nx,lags)

M = 200
Ufam,Pbt,Pb0t = covfamrun(ug0,Nx,lags,Bc_sq,M)

ind_lags_plot = range(0,lags+1,5)
N_lags_plot = np.size(ind_lags_plot)
lim = 2

for jpl in range(2):
 if jpl==0:
  Bplot = Bt;   Pbplot = Pbt
 if jpl==1:
  Bplot = B0t;  Pbplot = Pb0t  
 plt.figure() 
 my_cmap = matplotlib.cm.get_cmap('BrBG_r')
 for jlags in range(N_lags_plot):

  plt.subplot(3,N_lags_plot,1+jlags)
  plt.imshow(np.array(Bplot[:,:,jlags]),cmap=my_cmap,vmin=-lim,vmax=lim)    
  
  plt.subplot(3,N_lags_plot,N_lags_plot+1+jlags)
  plt.imshow(np.array(Pbplot[:,:,jlags]),cmap=my_cmap,vmin=-lim,vmax=lim)    
  
  plt.subplot(3,N_lags_plot,2*N_lags_plot+1+jlags)
  plt.imshow(np.array(Lxx*Pbplot[:,:,jlags]),cmap=my_cmap,vmin=-lim,vmax=lim)      
  
 del jlags 
del jpl


#%%
# ---------------------------------------------------------------------------
# 4.2. SC-4DEnVar
lam = 3; loctype = 1
Lxx = getlocmat(Nx,Nx,np.eye(Nx),lam,loctype) # get the localisation matrix
Lxy = getlocmat(Nx,Nx_obs,H,lam,loctype) # get the localisation matrix


# 3.3. 4DENVAR-state-sc
import inc4DenV; reload(inc4DenV)
from inc4DenV import envar
M = 8;  locenvar = 1; 
ua4Den,ub4Den,Uaen4Den,uaen4Den,UFr4Den = envar(ug0,t,x,R,invR,H,yobs,\
       period_obs,obsperwin,gridobs,Nx_obs,Bc_sq,lam,Lxx,Lxy,locenvar,M)
                        
    
plt.figure()
for jpl in range(Nx):
 plt.subplot(nrow,ncol,jpl+1)
 plt.plot(t,ut[:,jpl],'-k')
 plt.plot(t,ub4Den[:,jpl],'-b')
 plt.plot(t,ua4Den[:,jpl],'-r')
 plt.title('x['+str(jpl+1)+']')
 plt.grid(True)
del jpl
for jpl in range(Nx_obs):
 plt.subplot(nrow,ncol,loc_obs[jpl]+1)
 plt.autoscale(False) 
 plt.scatter(tobs,yobs[:,jpl],20,'r')
del jpl

rmseb4Den = np.empty((Nsteps,2));     rmsea4Den = np.empty((Nsteps,2));

for job in range(2):
 rmseb4Den[:,job] = rmse_spread(ut[:,locs[job].astype(int)], \
                np.squeeze(ub4Den[:,locs[job].astype(int)]),None,1)
 rmsea4Den[:,job] = rmse_spread(ut[:,locs[job].astype(int)], \
                np.squeeze(ua4Den[:,locs[job].astype(int)]),None,1)
del job

plt.figure()
for job in range(2):
 plt.subplot(1,2,job+1)
 plt.plot(t,rmseb4Den[:,job],'-c.')
 plt.plot(t,rmsea4Den[:,job],'-m.')
del job
















